"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const deep_equal_1 = __importDefault(require("deep-equal"));
const request_promise_1 = __importDefault(require("request-promise"));
const wbb_logger_1 = __importDefault(require("wbb-logger"));
const logger = new wbb_logger_1.default('atlas-api.ts');
class AtlasApi {
    constructor(params) {
        this.params = params;
        const { baseUrl, groupId, clusterName, publicKey, privateKey } = params;
        this.defaultOptions = {
            auth: {
                user: publicKey,
                pass: privateKey,
                sendImmediately: false,
            },
            json: true,
        };
        this.ftsUrl = `${baseUrl}/groups/${groupId}/clusters/${clusterName}/fts`;
    }
    async putCustomAnalyzers(analyzers) {
        const url = `${this.ftsUrl}/analyzers`;
        return (0, request_promise_1.default)(Object.assign(Object.assign({}, this.defaultOptions), { url, body: analyzers }));
    }
    async getSearchIndexes(db, collection) {
        const url = `${this.ftsUrl}/indexes/${db}/${collection}`;
        return (0, request_promise_1.default)(Object.assign(Object.assign({}, this.defaultOptions), { url }));
    }
    async createSearchIndex(db, collection, index) {
        this.logIndexChange(db, collection, 'Creating', index.name);
        const url = `${this.ftsUrl}/indexes`;
        return (0, request_promise_1.default)(Object.assign(Object.assign({}, this.defaultOptions), { method: 'POST', url, body: Object.assign({ collectionName: collection, database: db }, index) }));
    }
    async ensureSearchIndexes(db, collection, createParams) {
        const existingIndexes = await this.getSearchIndexes(db, collection);
        const toRemove = new Map();
        existingIndexes.forEach((e) => toRemove.set(e.name, e));
        for (const cParams of createParams) {
            const { name } = cParams;
            const existing = toRemove.get(name);
            if (existing) {
                toRemove.delete(name);
                await this.updateSearchIndex(db, collection, existing, cParams);
            }
            else {
                await this.createSearchIndex(db, collection, cParams);
            }
        }
        for (const index of toRemove.values()) {
            await this.deleteSearchIndex(db, collection, index);
        }
        return await this.getSearchIndexes(db, collection);
    }
    async updateSearchIndex(db, collection, existingIndex, createParams) {
        if ((0, deep_equal_1.default)(existingIndex.mappings, createParams.mappings)) {
            return;
        }
        this.logIndexChange(db, collection, 'Updating', createParams.name);
        const url = `${this.ftsUrl}/indexes/${existingIndex.indexID}`;
        return (0, request_promise_1.default)(Object.assign(Object.assign({}, this.defaultOptions), { method: 'PATCH', url, body: Object.assign({ collectionName: collection, database: db }, createParams) }));
    }
    async deleteSearchIndex(db, collection, existingIndex) {
        this.logIndexChange(db, collection, 'Deleting', existingIndex.name);
        const url = `${this.ftsUrl}/indexes/${existingIndex.indexID}`;
        return (0, request_promise_1.default)(Object.assign(Object.assign({}, this.defaultOptions), { method: 'DELETE', url }));
    }
    logIndexChange(db, collection, action, indexName) {
        logger.info(`${action} search index "${indexName}" for collection ${db}.${collection}`);
    }
}
exports.default = AtlasApi;
//# sourceMappingURL=atlas-api.js.map