"use strict";
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const clone_deep_1 = __importDefault(require("clone-deep"));
const mongodb_1 = require("mongodb");
const cursor_iterator_1 = __importDefault(require("./cursor-iterator"));
const nongo_1 = __importDefault(require("./nongo"));
const validator_1 = __importDefault(require("./validator"));
const DEFAULT_PATHS = ['_id', 'revision', 'history', 'diffs'];
function excludePathsFromDiff(diff, excludedPaths = DEFAULT_PATHS) {
    return diff.filter((d) => {
        if (!excludedPaths.includes(d.path[0])) {
            return true;
        }
    });
}
class Model {
    constructor(nongo, obj) {
        this.nongo = nongo;
        this.obj = obj;
        // {@code instanceof} can be a bit funny when importing classes across
        // modules so this is an explicit type guard we can use
        this.isNongoModel = true;
        // Set this.db if nongo given
        if (nongo) {
            this.db = nongo.db;
        }
        // If the _id is a string turn it into an {@code ObjectId}
        if (this.obj && typeof this.obj._id === 'string') {
            this.obj._id = new mongodb_1.ObjectId(this.obj._id);
        }
        this.collection = this.defineCollection();
        this.collectionRevision = // e.g. BlockRevision
            this.schemaVersioningEnabled()
                ? `${this.defineCollection()}${nongo_1.default.REVISION}`
                : null;
        this.collectionEnvironmentMapping = // e.g. BlockEnvironmentMapping
            this.schemaVersioningEnabled()
                ? `${this.defineCollection()}${nongo_1.default.ENVIRONMENT_MAPPING}`
                : null;
        this.name = this.constructor.name;
        this.dynamic = false;
        this.initStructure();
    }
    getCollectionRevision() {
        return this.collectionRevision;
    }
    schemaVersioningEnabled() {
        return false;
    }
    getMongoIndexes() {
        return [];
    }
    getAtlasSearchIndexes() {
        return [];
    }
    async validate() {
        const schema = this.nongo.schema[this.name];
        return new validator_1.default(schema, this).validate();
    }
    idString() {
        return this.obj._id.toHexString();
    }
    async getLatestRevisionNumber(query) {
        var _a;
        const revObj = await this.db
            .collection(this.collectionRevision)
            .findOne(query, { sort: { revision: -1 } });
        let revision = 1;
        if (revObj) {
            revision += revObj.revision;
        }
        return { revision, newestRevisionId: (_a = revObj === null || revObj === void 0 ? void 0 : revObj._id) === null || _a === void 0 ? void 0 : _a.toHexString() };
    }
    /*
     * Upsert the model into the database
     */
    async save() {
        // Set the object id if there isn't one already
        this.obj._id = nongo_1.default.toObjectId(this.obj._id);
        // Make a deep clone so that the original object does not have it's fields striped
        this.obj = (0, clone_deep_1.default)(this.obj);
        // Strip all the keys from the object that are not present in the schema
        // if not a dynamic model
        if (!this.dynamic) {
            this.strip('', this.obj);
            // Validate the object
            const err = await this.validate();
            if (err) {
                this.throwValidationError(err);
            }
        }
        if (this.collectionRevision) {
            const query = mongodb_1.ObjectId.isValid(this.obj.environment)
                ? { _id: new mongodb_1.ObjectId(this.obj.environment) }
                : { isLive: true };
            const clientEnv = await this.db.collection('Environment').findOne(query);
            const environmentId = clientEnv._id.toHexString();
            const { name, module, key } = this.obj;
            const revisionQuery = name ? { name, module } : { key };
            // Get id of newest revision, so we can use that to find the appropriate mapping
            // this is in case the object we're trying to save either doesn't have _id (for example if it's being imported)
            // or if the _id simply isn't correct for the environment
            const { revision, newestRevisionId } = await this.getLatestRevisionNumber(revisionQuery);
            this.obj.previousRevision = this.obj.revision;
            this.obj.revision = revision;
            // needed to create a new environment mapping
            let insertedRevisionId;
            if (clientEnv.isLive) {
                const liveRevisionObject = await this.db
                    .collection(this.collection)
                    .findOne(revisionQuery);
                // If object exists in live, override generated id
                this.obj._id = liveRevisionObject
                    ? liveRevisionObject._id
                    : this.obj._id;
                await this.db
                    .collection(this.collection)
                    .replaceOne({ _id: this.obj._id }, this.obj, { upsert: true });
                if (liveRevisionObject) {
                    // Check if the previously live version has a revision equivalent
                    const liveRevisionCopy = await this.db
                        .collection(this.collectionRevision)
                        .find(Object.assign(Object.assign({}, query), { revision: liveRevisionObject.revision }));
                    if (!liveRevisionCopy) {
                        liveRevisionObject._id = new mongodb_1.ObjectId();
                        await this.db
                            .collection(this.collectionRevision)
                            .insertOne(liveRevisionObject);
                    }
                }
                // Add new live model to revision, so they're both in sync
                const latestRevision = await this.db
                    .collection(this.collectionRevision)
                    .insertOne(Object.assign(Object.assign({}, this.obj), { _id: new mongodb_1.ObjectId() }));
                insertedRevisionId = latestRevision.insertedId.toHexString();
            }
            else {
                this.obj._id = new mongodb_1.ObjectId();
                const r = await this.db
                    .collection(this.collectionRevision)
                    .insertOne(this.obj);
                insertedRevisionId = r.insertedId.toHexString();
            }
            if (this.obj.revision !== 1) {
                // update relevant mappings
                await this.updateEnvironmentMapping(insertedRevisionId, newestRevisionId, environmentId, clientEnv.isLive);
            }
            if (this.obj.revision === 1) {
                // latest revision id, as that's what we just inserted, and need to save as baseline
                await this.createEnvironmentMapping(environmentId, insertedRevisionId, clientEnv.isLive);
            }
            if (this.obj.deleted) {
                await this.removeEnvironmentMapping(environmentId, newestRevisionId);
            }
        }
        else {
            // Upsert non revision models
            await this.db
                .collection(this.collection)
                .replaceOne({ _id: this.obj._id }, this.obj, { upsert: true });
        }
        // Return the saved model
        return this;
    }
    async removeEnvironmentMapping(environmentId, latestRevisionId) {
        const path = `environments.${environmentId}`;
        await this.db
            .collection(this.collectionEnvironmentMapping)
            .updateOne({ latestRevisionId }, { $unset: { [path]: '' } });
    }
    async createEnvironmentMapping(environmentId, latestRevisionId, isLive) {
        await this.db.collection(this.collectionEnvironmentMapping).insertOne({
            latestRevisionId: latestRevisionId,
            environments: {
                [environmentId]: isLive ? this.obj._id.toHexString() : latestRevisionId,
            },
        });
    }
    async updateEnvironmentMapping(insertedId, latestRevisionId, environmentId, isLive) {
        const path = `environments.${environmentId}`;
        // Find the mapping where the current
        await this.db.collection(this.collectionEnvironmentMapping).updateOne({ latestRevisionId }, {
            $set: {
                [path]: isLive ? this.obj._id.toHexString() : insertedId,
                latestRevisionId: insertedId,
            },
        });
    }
    async revertVersion(id, revisionId) {
        // get the actuall object
        const obj = await this.db
            .collection(this.collection)
            .findOne({ _id: new mongodb_1.ObjectId(id) });
        // get revision to revert to
        const objToRevert = await this.db
            .collection(this.collectionRevision)
            .findOne({ _id: new mongodb_1.ObjectId(revisionId) });
        // set id of revision to id of original, so it saves properly
        objToRevert._id = obj._id;
        // revert
        await this.db
            .collection(this.collection)
            .replaceOne({ _id: obj._id }, objToRevert);
        // Return the doc as a model
        // @ts-ignore
        return new this.constructor(this.nongo, objToRevert);
    }
    async getRevisionObjForEnvironment(isLive, id) {
        let doc;
        const _id = new mongodb_1.ObjectId(id);
        if (isLive) {
            doc = await this.db.collection(this.collection).findOne({ _id });
        }
        else {
            doc = await this.db.collection(this.collectionRevision).findOne({ _id });
        }
        // Return null if no doc was found
        if (doc == null) {
            return null;
        }
        // Return the doc as a model
        // @ts-ignore
        return new this.constructor(this.nongo, doc);
    }
    async saveEnvironmentRevision(id, environmentId) {
        const obj = await this.db
            .collection(this.collection)
            .findOne({ _id: new mongodb_1.ObjectId(id) });
        const revisionObj = (0, clone_deep_1.default)(obj);
        revisionObj._id = new mongodb_1.ObjectId();
        revisionObj.environment = environmentId;
        await this.db.collection(this.collectionRevision).insertOne(revisionObj);
        // Return the doc as a model
        // @ts-ignore
        return new this.constructor(this.nongo, revisionObj);
    }
    async removeEnvironmentRevisions(environmentId) {
        const result = await this.db
            .collection(this.collectionRevision)
            .deleteMany({ environment: environmentId });
        return result;
    }
    /*
     * Recursive method used to strip fields from an object have not been defined
     * by the schema.
     */
    strip(parentPath, toStrip) {
        for (const key of Object.keys(toStrip)) {
            const path = parentPath ? parentPath + '.' + key : key;
            // Skip over keys which shouldn't be stripped
            if (this.nongo.noStripKeys[this.constructor.name].has(path)) {
                continue;
            }
            // If the key is not in the schema delete it
            if (!this.nongo.validKeys[this.constructor.name].has(path)) {
                this.deleteKey(path, this.obj);
                // If the key is in the schema examin it's children if there are any
            }
            else {
                const value = toStrip[key];
                // If value is an array call strip on all of it's elements that are objects
                if (Array.isArray(value)) {
                    value.forEach((element) => {
                        if (typeof element === 'object') {
                            this.strip(path, element);
                        }
                    });
                    // If the value is a nested object call strip on it
                }
                else if (value != null && typeof value === 'object') {
                    this.strip(path, value);
                }
            }
        }
    }
    /*
     * Recursive method used to delete keys from models using a schema path.
     * This includeds the handling of arrays
     *
     * @param the path to the key which should be deleted.
     * @param the object to which the path is relative.
     */
    deleteKey(path, obj) {
        if (!obj) {
            return;
        }
        // If there is no . remaining in the path then delete the required key
        const indexFirstDot = path.indexOf('.');
        if (indexFirstDot === -1) {
            delete obj[path];
        }
        else {
            // Make a recusive call dependant on the value type
            const value = obj[path.slice(0, indexFirstDot)];
            const pathRemaining = path.slice(indexFirstDot + 1);
            if (Array.isArray(value)) {
                value.forEach((el) => this.deleteKey(pathRemaining, el));
            }
            else {
                this.deleteKey(pathRemaining, value);
            }
        }
    }
    async byId(id) {
        return this.findOne({ _id: id });
    }
    async getRevisions(query, options) {
        return this.toArray(await this.findIterator(query, options, this.collectionRevision));
    }
    async getRevision(query) {
        const doc = await this.db
            .collection(this.collectionRevision)
            .findOne(query);
        // Return null if no doc was found
        if (doc == null) {
            return null;
        }
        // Otherwise return the doc as a model
        // @ts-ignore
        return new this.constructor(this.nongo, doc);
    }
    /*
     * @param query: A mongo db query.
     * @param options: the query options e.g. sorting, limiting results etc.
     * @return the results of the query as an array of models.
     */
    async find(query = {}, options) {
        return this.toArray(await this.findIterator(query, options));
    }
    /*
     * @param operations: https://docs.mongodb.com/manual/core/bulk-write-operations/
     * @param options?: CollectionBulkWriteOptions
     */
    async bulkWrite(operations, options) {
        return this.db.collection(this.collection).bulkWrite(operations, options);
    }
    /*
     * @param query: A mongo db query.
     * @param options: the query options e.g. sorting, limiting results etc.
     * @return the results of the query as an {@link AsyncIterator}.
     */
    async findIterator(query = {}, options, collection = this.collection) {
        query = this.prepareQuery(query);
        const cursor = this.db.collection(collection).find(query, options);
        return new cursor_iterator_1.default(cursor, 
        // @ts-ignore
        (obj) => new this.constructor(this.nongo, obj));
    }
    /*
     * @param query: A mongo db query.
     * @return the result of the query as a model.
     */
    async findOne(query = {}) {
        query = this.prepareQuery(query);
        const doc = await this.db.collection(this.collection).findOne(query);
        // Return null if no doc was found
        if (doc == null) {
            return null;
        }
        // Otherwise return the doc as a model
        // @ts-ignore
        return new this.constructor(this.nongo, doc);
    }
    async remove(query, options) {
        query = this.prepareQuery(query);
        return this.db.collection(this.collection).deleteMany(query, options);
    }
    async aggregate(pipeline, options) {
        return this.toArray(await this.aggregateIterator(pipeline, options));
    }
    async aggregateIterator(pipeline, options) {
        options = options || {};
        if (options.allowDiskUse == null) {
            options.allowDiskUse = true;
        }
        const cursor = await this.db
            .collection(this.collection)
            .aggregate(pipeline, options);
        // @ts-ignore
        return new cursor_iterator_1.default(cursor);
    }
    async distinct(field, filter = {}) {
        return this.db.collection(this.collection).distinct(field, filter);
    }
    async update(query, updates, options) {
        query = this.prepareQuery(query);
        return this.db
            .collection(this.collection)
            .updateMany(query, updates, options);
    }
    async count(query = {}, options) {
        return this.db.collection(this.collection).countDocuments(query, options);
    }
    defineSchemaNoIndex() {
        const schema = this.defineSchema();
        this.removeSchemaIndex(schema);
        return schema;
    }
    /*
     * Must be run on all queries before they are used.
     */
    prepareQuery(query) {
        query = (0, clone_deep_1.default)(query);
        const { _id } = query;
        if (!_id) {
            return query;
        }
        // Make sure that _id values are ObjectIds not strings
        if (typeof _id === 'string') {
            query._id = nongo_1.default.toObjectId(query._id);
        }
        else if (_id.$in) {
            _id.$in = _id.$in.map((el) => typeof el === 'string' ? nongo_1.default.toObjectId(el) : el);
        }
        return query;
    }
    /*
     * This method should be overridden if want to use something other than the
     * class name as the collection name.
     */
    defineCollection() {
        return this.constructor.name;
    }
    throwValidationError(errors) {
        let objStr;
        try {
            objStr = `failed to save the following obj: ${JSON.stringify(this.obj)}`;
        }
        catch (err) {
            objStr = `failed to stringify obj: ${err.message}`;
        }
        throw Error(`${JSON.stringify(errors)} ${objStr}`);
    }
    /**
     * Recursively remove all the keys form the schema that are used to create indexes.
     *
     * @param parentSchema
     */
    removeSchemaIndex(parentSchema) {
        for (const key of Object.keys(parentSchema)) {
            // Delete any of the keys that can be used to create indexes
            const schema = parentSchema[key];
            delete schema.unique;
            delete schema.indexed;
            delete schema.text;
            // Make a recursive call if required
            const type = schema.type;
            if (Array.isArray(type)) {
                const elType = type[0];
                if (typeof elType === 'object') {
                    this.removeSchemaIndex(elType);
                }
            }
            else if (typeof type === 'object') {
                this.removeSchemaIndex(type);
            }
        }
    }
    async toArray(cursorIterator) {
        var _a, e_1, _b, _c;
        const arr = [];
        try {
            // @ts-ignore
            for (var _d = true, cursorIterator_1 = __asyncValues(cursorIterator), cursorIterator_1_1; cursorIterator_1_1 = await cursorIterator_1.next(), _a = cursorIterator_1_1.done, !_a;) {
                _c = cursorIterator_1_1.value;
                _d = false;
                try {
                    const m = _c;
                    arr.push(m);
                }
                finally {
                    _d = true;
                }
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (!_d && !_a && (_b = cursorIterator_1.return)) await _b.call(cursorIterator_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
        return arr;
    }
    /*
     * Initialises any objects and arrays which are marked as required in the schema.
     */
    initStructure() {
        // Init this.obj if it is null
        if (this.obj == null) {
            this.obj = {};
        }
        // We get the schema from nongo not this.getSchema() for the sake of efficiency
        if (this.nongo && this.nongo.schema[this.name]) {
            this.initKey(this.obj, this.nongo.schema[this.name]);
        }
    }
    initKey(parent, parentSchema) {
        for (const key of Object.keys(parentSchema)) {
            const schema = parentSchema[key];
            const type = schema.type;
            // Evaluate schema.required to a boolean
            let required = Array.isArray(schema.required)
                ? schema.required[0]
                : schema.required;
            required =
                typeof required === 'function'
                    ? required(parent[key], this.obj)
                    : required;
            const isNullRequired = parent[key] == null && required;
            if (Array.isArray(type)) {
                if (isNullRequired) {
                    parent[key] = [];
                }
                // Make recursive call if not null and an array and child type is not native type
                if (parent[key] != null && Array.isArray(parent[key])) {
                    const childType = type[0];
                    if (typeof childType === 'object') {
                        parent[key].forEach((el) => this.initKey(el, childType));
                    }
                }
            }
            else if (typeof type === 'object') {
                // Init the the object if null and required
                if (isNullRequired) {
                    parent[key] = {};
                }
                // Make recusive call if not null and an object
                if (parent[key] != null && typeof parent[key] === 'object') {
                    this.initKey(parent[key], type);
                }
            }
            else if (type === 'object' && isNullRequired) {
                parent[key] = {};
            }
        }
    }
}
exports.default = Model;
//# sourceMappingURL=model.js.map