"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const deep_equal_1 = __importDefault(require("deep-equal"));
const mongodb_1 = require("mongodb");
const wbb_logger_1 = __importDefault(require("wbb-logger"));
const inc_1 = __importDefault(require("semver/functions/inc"));
const atlas_api_1 = __importDefault(require("./atlas-api"));
const mongo_index_1 = __importDefault(require("./mongo-index"));
const mongo_index_set_1 = __importDefault(require("./mongo-index-set"));
const schema_parser_1 = __importDefault(require("./schema-parser"));
const logger = new wbb_logger_1.default('nongo.ts');
class Nongo {
    /*
     * @return a new ObjectId if {@code id} is null, an ObjectId created using {@code id}.
     */
    static toObjectId(id) {
        // Set the object id if there isn't one already
        if (id == null) {
            id = new mongodb_1.ObjectId();
            // If {@code this.obj._id} is not an {@code ObjectId} turn it into one
        }
        else if (!(id instanceof mongodb_1.ObjectId)) {
            id = new mongodb_1.ObjectId(id);
        }
        return id;
    }
    static revisionNumber(revision) {
        if (!revision) {
            return '0.0.1';
        }
        return (0, inc_1.default)(revision, 'patch');
    }
    // {@code modelClasses} should actually be of type {@code Model[]} throws a
    // compiler error when calling {@code Class.name}
    constructor(params, modelClasses, schemaParsers, mongoClient, poolSize) {
        this.params = params;
        this.modelClasses = modelClasses;
        this.schemaParsers = schemaParsers;
        this.mongoClient = mongoClient;
        this.poolSize = poolSize;
        this.schema = {};
        this.validKeys = {};
        this.noStripKeys = {};
        // Map of collection names to their relative instance of Model
        this.instanceMap = {};
        // Map of collection names to a function that when called returns a newly created instance of the appropriate Model
        this.newGeneratorMap = {};
    }
    async connect() {
        var _a;
        // Connect to the database
        this.mongoClient =
            (_a = this.mongoClient) !== null && _a !== void 0 ? _a : (await new mongodb_1.MongoClient(this.uri(), {
                minPoolSize: this.poolSize,
            }).connect());
        this.db = this.mongoClient.db(this.params.db);
        this.dbName = this.params.db;
        // Map the classes to instances
        this.instances = this.modelClasses.map((Clazz) => new Clazz(this, null));
        // Init the schema parsers if they are aren't already given
        if (!this.schemaParsers) {
            this.schemaParsers = this.instances.map((instance) => {
                const collectionName = instance.name;
                // Validate the schema (throws error if not valid)
                return new schema_parser_1.default(instance.defineSchema(), collectionName);
            });
        }
        const collectionNames = await this.getCollectionNames();
        // Iterate over each of the model classes
        for (let i = 0; i < this.instances.length; i++) {
            const Clazz = this.modelClasses[i];
            const instance = this.instances[i];
            const parser = this.schemaParsers[i];
            const schema = parser.schema;
            // Create a field on {@code this} that can be used to access what appear to
            // client code to be static methods of the model e.g. {@code nongo.MyModel.find(..)}
            this.instanceMap[instance.name] = instance;
            // Create a field on {@code this.validKeys} for the model
            this.validKeys[instance.name] = parser.validKeys;
            this.noStripKeys[instance.name] = parser.noStripKeys;
            // Create a field on {@code this.schema} for the model
            this.schema[instance.name] = schema;
            // Create the collection if it doesn't exist
            const collName = instance.collection;
            if (!collectionNames.includes(collName)) {
                await this.db.createCollection(collName);
            }
            // Create the revision collection if it doesn't exist
            const revisionCollName = instance.getCollectionRevision();
            if (revisionCollName && !collectionNames.includes(revisionCollName)) {
                await this.db.createCollection(revisionCollName);
            }
            // Add field to {@code this.newGenerator} with a function as a value that can be
            // used to init a new model
            this.newGeneratorMap[instance.name] = (obj) => {
                return new Clazz(this, obj);
            };
        }
        return this;
    }
    async ensureIndexes() {
        // collections in Mongo
        const collectionNames = await this.getCollectionNames();
        const indexedCollections = [];
        // Iterate over each of the model instances we have so we can ensure their indexes
        for (let i = 0; i < this.instances.length; i++) {
            const instance = this.instances[i];
            const parser = this.schemaParsers[i];
            const schema = parser.schema;
            // Ensure the indexes for the collection
            const explicitIndexes = instance.getMongoIndexes();
            const collName = instance.collection;
            try {
                await this.ensureCollectionIndexes(schema, collName, explicitIndexes, collectionNames);
            }
            catch (err) {
                logger.error(`Failed to ensure indexes for ${collName}`);
                logger.error(err);
            }
            // Make sure we have atlas params if we need them
            const searchIndexes = instance.getAtlasSearchIndexes();
            const atlasParams = this.params.atlas;
            const atlasEnabled = atlasParams && !atlasParams.disabled;
            if (atlasEnabled && searchIndexes.length) {
                await new atlas_api_1.default(atlasParams).ensureSearchIndexes(this.dbName, instance.name, searchIndexes);
            }
            // If the instance we've just indexed has a revision,
            // add to list, so we don't drop it
            indexedCollections.push(collName);
            const revision = instance.getCollectionRevision();
            if (revision) {
                indexedCollections.push(revision);
            }
        }
        const dropCollections = collectionNames.filter((c) => !indexedCollections.includes(c));
        for (const coll of dropCollections) {
            logger.info(`Dropping ${this.dbName}.${coll}`);
            await this.dropCollection(coll);
        }
    }
    async dropCollection(collName) {
        await this.db.dropCollection(collName);
    }
    async dropDatabase() {
        await this.db.dropDatabase();
    }
    /**
     * Remove all documents in the database
     */
    async removeAll() {
        for (const instance of this.instances || []) {
            const revisionCollection = instance.getCollectionRevision();
            if (revisionCollection) {
                this.db.collection(revisionCollection).deleteMany({});
            }
            await instance.db.collection(instance.collection).deleteMany({});
        }
    }
    async dbExists() {
        const allDbs = (await this.db.admin().listDatabases()).databases;
        return allDbs.some((db) => db.name === this.dbName);
    }
    /**
     * Used to return the "collection instance" of a {@link Model}. You should use this method to gain access to an object
     * which you can use to run queries again a collection such as {Model.find(..)} etc.
     *
     * @param {{new(...args: any[]): T}} model
     * @returns {T}
     */
    col(model) {
        return this.instanceMap[model.name];
    }
    /**
     * Used to create a new instance of a {@link Model}
     *
     * @param {{new(...args: any[]): T}} model
     * @param {T["obj"]} obj
     * @returns {T}
     */
    new(model, obj) {
        return this.newGeneratorMap[model.name](obj);
    }
    /**
     * @returns {string} the mongo uri built from {@code this.params}.
     */
    uri() {
        // If given explicitly just return it
        if (this.params.uri) {
            return this.params.uri;
        }
        // Protocol
        let url = (this.params.protocol || 'mongodb') + '://';
        // Authentication (optional)
        if (this.params.username) {
            url += this.params.username + ':' + this.params.password + '@';
        }
        // Host
        url += this.params.host;
        // Then optionally the port
        // Port (optional)
        if (this.params.port) {
            url += ':' + this.params.port;
        }
        // Database
        url += '/' + this.params.db + '?';
        // Options given as uri params
        const options = this.params.options || {};
        url += Object.keys(options)
            .map((key) => key + '=' + encodeURI(options[key]))
            .join('&');
        return url;
    }
    async getCollectionNames() {
        const collections = await this.db.listCollections().toArray();
        return collections.map((col) => col.name);
    }
    async ensureCollectionIndexes(schema, collName, explicitIndexes, collectionNames) {
        const context = this.dbName + '.' + collName;
        // Get the existing set of indexes, we will remove the ones we want to keep
        const existingIndexes = await this.existingIndexes(collName, collectionNames);
        // Ensure all the required indexes have been applied
        const indexSet = await this.getSchemaIndexes(schema);
        explicitIndexes.forEach((idx) => indexSet.add(idx));
        const collection = this.db.collection(collName);
        // Remove any indexes that are no longer required
        for (const index of existingIndexes.values()) {
            if (!indexSet.get(index)) {
                logger.info(`${context} dropping ${index} as it is no longer required...`);
                try {
                    await collection.dropIndex(index.name);
                }
                catch (e) {
                    logger.error(e.message);
                }
            }
        }
        for (const index of indexSet.values()) {
            // Always background indexes
            index.options.background = true;
            // Get the existing index index with the same key if there is one
            const existingIndex = existingIndexes.get(index);
            // If there isn't log that we are creating a new one
            if (!existingIndex) {
                logger.info(`${context} is creating ${index}...`);
            }
            // If there is an existing index with same name and different config, drop it
            const replacementRequired = existingIndex && !(0, deep_equal_1.default)(index, existingIndex);
            if (replacementRequired) {
                logger.info(`${context} is replacing ${existingIndex} with ${index}...`);
                await collection.dropIndex(existingIndex.name);
            }
            // Ensure the index
            await collection.createIndex(index.index, index.options);
            // Wait for the background indexing to be complete
            await this.waitForIndex(collection, index);
            // If we have built an index do some logging
            if (replacementRequired || !existingIndex) {
                logger.info(`${context} finished building ${index}`);
            }
        }
    }
    async waitForIndex(collection, index) {
        let indexingCompete = false;
        while (indexingCompete === false) {
            const indexNames = (await collection.indexInformation({ full: true, session: null })).map((idx) => idx.name);
            indexingCompete = indexNames.includes(index.name);
            // If not done indexes wait for a second to check again
            if (!indexingCompete) {
                await new Promise((resolve) => setTimeout(resolve, 1000));
            }
        }
    }
    async existingIndexes(collName, collNames) {
        const indexSet = new mongo_index_set_1.default();
        // If the collection exists
        if (collNames.includes(collName)) {
            // Iterate over the existing indexes
            const collection = this.db.collection(collName);
            for (const indexInfo of await collection.indexes()) {
                // Add all the indexes to the {@code indexSet} except _id
                const index = indexInfo.key;
                if (!(0, deep_equal_1.default)(index, { _id: 1 })) {
                    // Parse the index info
                    const options = {
                        unique: indexInfo.unique,
                        background: indexInfo.background,
                        name: indexInfo.name,
                        weights: indexInfo.weights,
                        sparse: indexInfo.sparse,
                    };
                    // if it's text index, retrieve the field names info from weights
                    if (indexInfo.textIndexVersion) {
                        const textIndex = {};
                        for (const fieldName of Object.keys(indexInfo.weights)) {
                            textIndex[fieldName] = 'text';
                        }
                        indexSet.add(new mongo_index_1.default(textIndex, options, false));
                    }
                    else {
                        indexSet.add(new mongo_index_1.default(index, options, false));
                    }
                }
            }
        }
        return indexSet;
    }
    /**
     * Recursively traverse the schema to populate the {@code indexList} with {@link MongoIndex}s constructed using schema
     * kv-pairs.
     *
     * @param parentSchema
     * @param {string} parentPath
     * @param indexSet
     * @returns {Promise<MongoIndex[]>}
     */
    async getSchemaIndexes(parentSchema, parentPath = '', indexSet = new mongo_index_set_1.default()) {
        for (const key of Object.keys(parentSchema)) {
            // Define the index
            const path = parentPath ? parentPath + '.' + key : key;
            // Create an index and add it to the index list if required. Indexes default to background
            const schema = parentSchema[key];
            if (schema.unique || schema.indexed || schema.text) {
                const index = {};
                index[path] = schema.text ? 'text' : 1;
                indexSet.add(new mongo_index_1.default(index, { background: true, unique: schema.unique }));
            }
            // Make a recursive call if required
            const type = schema.type;
            if (Array.isArray(type)) {
                const elType = type[0];
                if (typeof elType === 'object') {
                    await this.getSchemaIndexes(elType, path, indexSet);
                }
            }
            else if (typeof type === 'object') {
                await this.getSchemaIndexes(type, path, indexSet);
            }
        }
        return indexSet;
    }
}
exports.default = Nongo;
Nongo.REVISION = 'Revision';
Nongo.ENVIRONMENT_MAPPING = 'EnvironmentMapping';
//# sourceMappingURL=nongo.js.map