import deepEqual from 'deep-equal';
import rp from 'request-promise';
import Logger from 'wbb-logger';
import {Analyzer, AtlasParams, AtlasSearchIndex, CreateSearchIndex} from './interface/atlas';

const logger = new Logger('atlas-api.ts');

export default class AtlasApi {

  private readonly ftsUrl: string;
  private readonly defaultOptions;

  constructor(private params: AtlasParams) {
    const {baseUrl, groupId, clusterName, publicKey, privateKey} = params;

    this.defaultOptions = {
      auth: {
        user: publicKey,
        pass: privateKey,
        sendImmediately: false,
      },
      json: true,
    };

    this.ftsUrl = `${baseUrl}/groups/${groupId}/clusters/${clusterName}/fts`;
  }

  public async putCustomAnalyzers(analyzers: Analyzer[]): Promise<Analyzer[]> {
    const url = `${this.ftsUrl}/analyzers`;
    return rp({
      ...this.defaultOptions,
      url,
      body: analyzers,
    });
  }

  public async getSearchIndexes(db: string, collection: string): Promise<AtlasSearchIndex[]> {
    const url = `${this.ftsUrl}/indexes/${db}/${collection}`;
    return rp({
      ...this.defaultOptions,
      url,
    });
  }

  public async createSearchIndex(db: string, collection: string, index: CreateSearchIndex): Promise<any> {
    this.logIndexChange(db, collection, 'Creating', index.name);
    const url = `${this.ftsUrl}/indexes`;
    return rp({
      ...this.defaultOptions,
      method: 'POST',
      url,
      body: {
        collectionName: collection,
        database: db,
        ...index,
      },
    });
  }

  public async ensureSearchIndexes(db: string, collection: string, createParams: CreateSearchIndex[]): Promise<AtlasSearchIndex[]> {
    const existingIndexes = await this.getSearchIndexes(db, collection);
    const toRemove = new Map<string, AtlasSearchIndex>();
    existingIndexes.forEach((e) => toRemove.set(e.name, e));

    for (const cParams of createParams) {
      const {name} = cParams;
      const existing = toRemove.get(name);

      if (existing) {
        toRemove.delete(name);
        await this.updateSearchIndex(db, collection, existing, cParams);
      } else {
        await this.createSearchIndex(db, collection, cParams);
      }

    }

    for (const index of toRemove.values()) {
      await this.deleteSearchIndex(db, collection, index);
    }

    return await this.getSearchIndexes(db, collection);
  }

  public async updateSearchIndex(db: string, collection: string, existingIndex: AtlasSearchIndex, createParams: CreateSearchIndex): Promise<any> {
    if (deepEqual(existingIndex.mappings, createParams.mappings)) {
      return;
    }

    this.logIndexChange(db, collection, 'Updating', createParams.name);

    const url = `${this.ftsUrl}/indexes/${existingIndex.indexID}`;

    return rp({
      ...this.defaultOptions,
      method: 'PATCH',
      url,
      body: {
        collectionName: collection,
        database: db,
        ...createParams,
      },
    });
  }

  public async deleteSearchIndex(db: string, collection: string, existingIndex: AtlasSearchIndex): Promise<any> {
    this.logIndexChange(db, collection, 'Deleting', existingIndex.name);
    const url = `${this.ftsUrl}/indexes/${existingIndex.indexID}`;
    return rp({
      ...this.defaultOptions,
      method: 'DELETE',
      url,
    });
  }

  private logIndexChange(db: string, collection: string, action: string, indexName: string) {
    logger.info(`${action} search index "${indexName}" for collection ${db}.${collection}`);
  }

}
