import clone from 'clone-deep';
import {
  AggregateOptions,
  AnyBulkWriteOperation,
  BulkWriteOptions,
  BulkWriteResult,
  CountDocumentsOptions,
  Db,
  DeleteOptions,
  DeleteResult,
  Document,
  Filter,
  FindOptions,
  ObjectId,
  UpdateFilter,
  UpdateOptions,
  UpdateResult,
} from 'mongodb';

import CursorIterator from './cursor-iterator';
import {CreateSearchIndex} from './interface';
import MongoIndex from './mongo-index';
import Nongo from './nongo';
import Validator from './validator';

const DEFAULT_PATHS = ['_id', 'revision', 'history', 'diffs'];

function excludePathsFromDiff(diff, excludedPaths = DEFAULT_PATHS) {
  return diff.filter((d) => {
    if (!excludedPaths.includes(d.path[0])) {
      return true;
    }
  });
}


export default abstract class Model<
  T extends {
    _id?: ObjectId;
    revision?: number;
    previousRevision?: number;
    diffs?: any;
    name?: string;
    key?: string;
    environment?: string;
    module?: string;
    deleted?: true;
  } = any,
> {
  // {@code instanceof} can be a bit funny when importing classes across
  // modules so this is an explicit type guard we can use
  public isNongoModel: true = true;
  public collection: string;
  public db: Db;
  public name: string;
  public dynamic: boolean;
  private collectionRevision: string | null;
  private collectionEnvironmentMapping: string | null;

  constructor(public nongo: Nongo, public obj: T) {
    // Set this.db if nongo given
    if (nongo) {
      this.db = nongo.db;
    }

    // If the _id is a string turn it into an {@code ObjectId}
    if (this.obj && typeof this.obj._id === 'string') {
      this.obj._id = new ObjectId(this.obj._id);
    }

    this.collection = this.defineCollection();
    this.collectionRevision = // e.g. BlockRevision
      this.schemaVersioningEnabled()
        ? `${this.defineCollection()}${Nongo.REVISION}`
        : null;
    this.collectionEnvironmentMapping = // e.g. BlockEnvironmentMapping
      this.schemaVersioningEnabled()
        ? `${this.defineCollection()}${Nongo.ENVIRONMENT_MAPPING}`
        : null;
    this.name = this.constructor.name;
    this.dynamic = false;
    this.initStructure();
  }

  public getCollectionRevision(): string {
    return this.collectionRevision;
  }

  protected schemaVersioningEnabled(): boolean {
    return false;
  }

  public getMongoIndexes(): MongoIndex[] {
    return [];
  }

  public getAtlasSearchIndexes(): CreateSearchIndex[] {
    return [];
  }

  public async validate() {
    const schema = this.nongo.schema[this.name];
    return new Validator(schema, this).validate();
  }

  public idString(): string {
    return this.obj._id.toHexString();
  }

  private async getLatestRevisionNumber(
    query,
  ): Promise<{revision: number; newestRevisionId: string}> {
    const revObj = await this.db
      .collection(this.collectionRevision)
      .findOne(query, {sort: {revision: -1}});
    let revision = 1;
    if (revObj) {
      revision += revObj.revision;
    }
    return {revision, newestRevisionId: revObj?._id?.toHexString()};
  }

  /*
   * Upsert the model into the database
   */
  public async save() {
    // Set the object id if there isn't one already
    this.obj._id = Nongo.toObjectId(this.obj._id);

    // Make a deep clone so that the original object does not have it's fields striped
    this.obj = clone(this.obj);
    // Strip all the keys from the object that are not present in the schema
    // if not a dynamic model
    if (!this.dynamic) {
      this.strip('', this.obj);
      // Validate the object
      const err = await this.validate();
      if (err) {
        this.throwValidationError(err);
      }
    }

    if (this.collectionRevision) {
      const query = ObjectId.isValid(this.obj.environment)
        ? {_id: new ObjectId(this.obj.environment)}
        : {isLive: true};
      const clientEnv = await this.db.collection('Environment').findOne(query);
      const environmentId = clientEnv._id.toHexString();

      const {name, module, key} = this.obj;
      const revisionQuery = name ? {name, module} : {key};
      // Get id of newest revision, so we can use that to find the appropriate mapping
      // this is in case the object we're trying to save either doesn't have _id (for example if it's being imported)
      // or if the _id simply isn't correct for the environment
      const {revision, newestRevisionId} = await this.getLatestRevisionNumber(
        revisionQuery,
      );
      this.obj.previousRevision = this.obj.revision;
      this.obj.revision = revision;
      // needed to create a new environment mapping
      let insertedRevisionId: string;
      if (clientEnv.isLive) {
        const liveRevisionObject = await this.db
          .collection(this.collection)
          .findOne(revisionQuery);
        // If object exists in live, override generated id
        this.obj._id = liveRevisionObject
          ? liveRevisionObject._id
          : this.obj._id;
        await this.db
          .collection(this.collection)
          .replaceOne({_id: this.obj._id}, this.obj, {upsert: true});
        if (liveRevisionObject) {
          // Check if the previously live version has a revision equivalent
          const liveRevisionCopy = await this.db
            .collection(this.collectionRevision)
            .find({...query, revision: liveRevisionObject.revision});
          if (!liveRevisionCopy) {
            liveRevisionObject._id = new ObjectId();
            await this.db
              .collection(this.collectionRevision)
              .insertOne(liveRevisionObject);
          }
        }
        // Add new live model to revision, so they're both in sync
        const latestRevision = await this.db
          .collection(this.collectionRevision)
          .insertOne({...this.obj, _id: new ObjectId()});
        insertedRevisionId = latestRevision.insertedId.toHexString();
      } else {
        this.obj._id = new ObjectId();
        const r = await this.db
          .collection(this.collectionRevision)
          .insertOne(this.obj);
        insertedRevisionId = r.insertedId.toHexString();
      }
      if (this.obj.revision !== 1) {
        // update relevant mappings
        await this.updateEnvironmentMapping(
          insertedRevisionId,
          newestRevisionId,
          environmentId,
          clientEnv.isLive,
        );
      }
      if (this.obj.revision === 1) {
        // latest revision id, as that's what we just inserted, and need to save as baseline
        await this.createEnvironmentMapping(
          environmentId,
          insertedRevisionId,
          clientEnv.isLive,
        );
      }
      if (this.obj.deleted) {
        await this.removeEnvironmentMapping(environmentId, newestRevisionId);
      }
    } else {
      // Upsert non revision models
      await this.db
        .collection(this.collection)
        .replaceOne({_id: this.obj._id}, this.obj, {upsert: true});
    }

    // Return the saved model
    return this;
  }

  private async removeEnvironmentMapping(
    environmentId: string,
    latestRevisionId: string,
  ): Promise<void> {
    const path = `environments.${environmentId}`;
    await this.db
      .collection(this.collectionEnvironmentMapping)
      .updateOne({latestRevisionId}, {$unset: {[path]: ''}});
  }

  private async createEnvironmentMapping(
    environmentId: string,
    latestRevisionId: string,
    isLive: boolean,
  ): Promise<void> {
    await this.db.collection(this.collectionEnvironmentMapping).insertOne({
      latestRevisionId: latestRevisionId,
      environments: {
        [environmentId]: isLive ? this.obj._id.toHexString() : latestRevisionId,
      },
    });
  }

  private async updateEnvironmentMapping(
    insertedId: string,
    latestRevisionId: string,
    environmentId: string,
    isLive: boolean,
  ): Promise<void> {
    const path = `environments.${environmentId}`;
    // Find the mapping where the current
    await this.db.collection(this.collectionEnvironmentMapping).updateOne(
      {latestRevisionId},
      {
        $set: {
          [path]: isLive ? this.obj._id.toHexString() : insertedId,
          latestRevisionId: insertedId,
        },
      },
    );
  }

  public async revertVersion(id: string, revisionId: string): Promise<this> {
    // get the actuall object
    const obj = await this.db
      .collection(this.collection)
      .findOne({_id: new ObjectId(id)});

    // get revision to revert to
    const objToRevert = await this.db
      .collection(this.collectionRevision)
      .findOne({_id: new ObjectId(revisionId)});

    // set id of revision to id of original, so it saves properly
    objToRevert._id = obj._id;

    // revert
    await this.db
      .collection(this.collection)
      .replaceOne({_id: obj._id}, objToRevert);

    // Return the doc as a model
    // @ts-ignore
    return new this.constructor(this.nongo, objToRevert);
  }

  public async getRevisionObjForEnvironment(
    isLive: boolean,
    id: string,
  ): Promise<this> {
    let doc;
    const _id = new ObjectId(id);

    if (isLive) {
      doc = await this.db.collection(this.collection).findOne({_id});
    } else {
      doc = await this.db.collection(this.collectionRevision).findOne({_id});
    }

    // Return null if no doc was found
    if (doc == null) {
      return null;
    }

    // Return the doc as a model
    // @ts-ignore
    return new this.constructor(this.nongo, doc);
  }

  public async saveEnvironmentRevision(
    id: string,
    environmentId: string,
  ): Promise<this> {
    const obj = await this.db
      .collection(this.collection)
      .findOne({_id: new ObjectId(id)});

    const revisionObj = clone(obj);
    revisionObj._id = new ObjectId();
    revisionObj.environment = environmentId;

    await this.db.collection(this.collectionRevision).insertOne(revisionObj);
    // Return the doc as a model
    // @ts-ignore
    return new this.constructor(this.nongo, revisionObj);
  }

  public async removeEnvironmentRevisions(
    environmentId: string,
  ): Promise<DeleteResult> {
    const result = await this.db
      .collection(this.collectionRevision)
      .deleteMany({environment: environmentId});
    return result;
  }

  /*
   * Recursive method used to strip fields from an object have not been defined
   * by the schema.
   */
  public strip(parentPath, toStrip) {
    for (const key of Object.keys(toStrip)) {
      const path = parentPath ? parentPath + '.' + key : key;

      // Skip over keys which shouldn't be stripped
      if (this.nongo.noStripKeys[this.constructor.name].has(path)) {
        continue;
      }

      // If the key is not in the schema delete it
      if (!this.nongo.validKeys[this.constructor.name].has(path)) {
        this.deleteKey(path, this.obj);

        // If the key is in the schema examin it's children if there are any
      } else {
        const value = toStrip[key];

        // If value is an array call strip on all of it's elements that are objects
        if (Array.isArray(value)) {
          value.forEach((element) => {
            if (typeof element === 'object') {
              this.strip(path, element);
            }
          });

          // If the value is a nested object call strip on it
        } else if (value != null && typeof value === 'object') {
          this.strip(path, value);
        }
      }
    }
  }

  /*
   * Recursive method used to delete keys from models using a schema path.
   * This includeds the handling of arrays
   *
   * @param the path to the key which should be deleted.
   * @param the object to which the path is relative.
   */
  public deleteKey(path, obj) {
    if (!obj) {
      return;
    }

    // If there is no . remaining in the path then delete the required key
    const indexFirstDot = path.indexOf('.');
    if (indexFirstDot === -1) {
      delete obj[path];
    } else {
      // Make a recusive call dependant on the value type
      const value = obj[path.slice(0, indexFirstDot)];
      const pathRemaining = path.slice(indexFirstDot + 1);
      if (Array.isArray(value)) {
        value.forEach((el) => this.deleteKey(pathRemaining, el));
      } else {
        this.deleteKey(pathRemaining, value);
      }
    }
  }

  public async byId(id) {
    return this.findOne({_id: id});
  }

  public async getRevisions(
    query: Filter<Document>,
    options?: FindOptions<Document>,
  ): Promise<this[]> {
    return this.toArray(
      await this.findIterator(query, options, this.collectionRevision),
    );
  }

  public async getRevision(query: Filter<Document>): Promise<this> {
    const doc = await this.db
      .collection(this.collectionRevision)
      .findOne(query);

    // Return null if no doc was found
    if (doc == null) {
      return null;
    }

    // Otherwise return the doc as a model
    // @ts-ignore
    return new this.constructor(this.nongo, doc);
  }

  /*
   * @param query: A mongo db query.
   * @param options: the query options e.g. sorting, limiting results etc.
   * @return the results of the query as an array of models.
   */
  public async find(
    query: Filter<Document> = {},
    options?: FindOptions<Document>,
  ): Promise<this[]> {
    return this.toArray(await this.findIterator(query, options));
  }

  /*
   * @param operations: https://docs.mongodb.com/manual/core/bulk-write-operations/
   * @param options?: CollectionBulkWriteOptions
   */
  public async bulkWrite(
    operations: AnyBulkWriteOperation<Document>[],
    options?: BulkWriteOptions,
  ): Promise<BulkWriteResult> {
    return this.db.collection(this.collection).bulkWrite(operations, options);
  }

  /*
   * @param query: A mongo db query.
   * @param options: the query options e.g. sorting, limiting results etc.
   * @return the results of the query as an {@link AsyncIterator}.
   */
  public async findIterator(
    query: Filter<Document> = {},
    options?: FindOptions<Document>,
    collection = this.collection,
  ): Promise<CursorIterator<this>> {
    query = this.prepareQuery(query);
    const cursor = this.db.collection(collection).find(query, options);
    return new CursorIterator<this>(
      cursor,
      // @ts-ignore
      (obj) => new this.constructor(this.nongo, obj),
    );
  }

  /*
   * @param query: A mongo db query.
   * @return the result of the query as a model.
   */
  public async findOne(query: Filter<Document> = {}): Promise<this> {
    query = this.prepareQuery(query);
    const doc = await this.db.collection(this.collection).findOne(query);

    // Return null if no doc was found
    if (doc == null) {
      return null;
    }

    // Otherwise return the doc as a model
    // @ts-ignore
    return new this.constructor(this.nongo, doc);
  }

  public async remove(
    query: Filter<Document>,
    options?: DeleteOptions,
  ): Promise<DeleteResult> {
    query = this.prepareQuery(query);
    return this.db.collection(this.collection).deleteMany(query, options);
  }

  public async aggregate(
    pipeline: Document[],
    options?: AggregateOptions,
  ): Promise<any[]> {
    return this.toArray(await this.aggregateIterator(pipeline, options));
  }

  public async aggregateIterator(
    pipeline: Document[],
    options?: AggregateOptions,
  ): Promise<CursorIterator<any>> {
    options = options || {};
    if (options.allowDiskUse == null) {
      options.allowDiskUse = true;
    }
    const cursor = await this.db
      .collection(this.collection)
      .aggregate(pipeline, options);
    // @ts-ignore
    return new CursorIterator(cursor);
  }

  public async distinct(field: string, filter: Filter<Document> = {}) {
    return this.db.collection(this.collection).distinct(field, filter);
  }

  public async update(
    query: Filter<Document>,
    updates: UpdateFilter<Document>,
    options?: UpdateOptions,
  ): Promise<UpdateResult<Document>> {
    query = this.prepareQuery(query);
    return this.db
      .collection(this.collection)
      .updateMany(query, updates, options);
  }

  public async count(
    query: Document = {},
    options?: CountDocumentsOptions,
  ): Promise<number> {
    return this.db.collection(this.collection).countDocuments(query, options);
  }

  /*
   * Return the schema for this model. See getSchema for an understanding of inheritance.
   */
  public abstract defineSchema(): any;

  public defineSchemaNoIndex(): any {
    const schema = this.defineSchema();
    this.removeSchemaIndex(schema);
    return schema;
  }

  /*
   * Must be run on all queries before they are used.
   */
  public prepareQuery(query) {
    query = clone(query);
    const {_id} = query;
    if (!_id) {
      return query;
    }

    // Make sure that _id values are ObjectIds not strings
    if (typeof _id === 'string') {
      query._id = Nongo.toObjectId(query._id);
    } else if (_id.$in) {
      _id.$in = _id.$in.map((el) =>
        typeof el === 'string' ? Nongo.toObjectId(el) : el,
      );
    }

    return query;
  }

  /*
   * This method should be overridden if want to use something other than the
   * class name as the collection name.
   */
  protected defineCollection() {
    return this.constructor.name;
  }

  private throwValidationError(errors: string[]) {
    let objStr;
    try {
      objStr = `failed to save the following obj: ${JSON.stringify(this.obj)}`;
    } catch (err) {
      objStr = `failed to stringify obj: ${err.message}`;
    }
    throw Error(`${JSON.stringify(errors)} ${objStr}`);
  }

  /**
   * Recursively remove all the keys form the schema that are used to create indexes.
   *
   * @param parentSchema
   */
  private removeSchemaIndex(parentSchema: any) {
    for (const key of Object.keys(parentSchema)) {
      // Delete any of the keys that can be used to create indexes
      const schema = parentSchema[key];
      delete schema.unique;
      delete schema.indexed;
      delete schema.text;

      // Make a recursive call if required
      const type = schema.type;
      if (Array.isArray(type)) {
        const elType = type[0];
        if (typeof elType === 'object') {
          this.removeSchemaIndex(elType);
        }
      } else if (typeof type === 'object') {
        this.removeSchemaIndex(type);
      }
    }
  }

  private async toArray(cursorIterator: CursorIterator<this>): Promise<this[]> {
    const arr = [];
    // @ts-ignore
    for await (const m of cursorIterator) {
      arr.push(m);
    }
    return arr;
  }

  /*
   * Initialises any objects and arrays which are marked as required in the schema.
   */
  private initStructure() {
    // Init this.obj if it is null
    if (this.obj == null) {
      this.obj = {} as T;
    }

    // We get the schema from nongo not this.getSchema() for the sake of efficiency
    if (this.nongo && this.nongo.schema[this.name]) {
      this.initKey(this.obj, this.nongo.schema[this.name]);
    }
  }

  private initKey(parent, parentSchema) {
    for (const key of Object.keys(parentSchema)) {
      const schema = parentSchema[key];
      const type = schema.type;

      // Evaluate schema.required to a boolean
      let required = Array.isArray(schema.required)
        ? schema.required[0]
        : schema.required;
      required =
        typeof required === 'function'
          ? required(parent[key], this.obj)
          : required;

      const isNullRequired = parent[key] == null && required;

      if (Array.isArray(type)) {
        if (isNullRequired) {
          parent[key] = [];
        }

        // Make recursive call if not null and an array and child type is not native type
        if (parent[key] != null && Array.isArray(parent[key])) {
          const childType = type[0];
          if (typeof childType === 'object') {
            parent[key].forEach((el) => this.initKey(el, childType));
          }
        }
      } else if (typeof type === 'object') {
        // Init the the object if null and required
        if (isNullRequired) {
          parent[key] = {};
        }

        // Make recusive call if not null and an object
        if (parent[key] != null && typeof parent[key] === 'object') {
          this.initKey(parent[key], type);
        }
      } else if (type === 'object' && isNullRequired) {
        parent[key] = {};
      }
    }
  }
}
